@@ -3,20 +3,24 @@ use warnings;
 use strict;
 use utf8;
 
-our $VERSION = '0.15';
+our $VERSION = '0.16';
 
 # Core modules
-use Carp qw/carp croak/;
+use Carp qw/carp croak confess cluck/;
 use Math::Trig qw!acos pi rad2deg deg2rad!;
 use Scalar::Util 'looks_like_number';
 
-# Installed modules
+# Modules the user needs to install
+
 use XML::Parser;
 use Cairo;
 use Image::SVG::Path qw/extract_path_info/;
 
 our $default_surface_type = 'argb32';
 our $default_surface_size = 100;
+
+# This is what the SVG standard says the default colour is.
+
 our @defaultrgb = (0, 0, 0);
 
 sub new
@@ -45,7 +49,7 @@ sub new
     if ($context) {
 	$self->{cr} = $context;
 	if ($surface) {
-	    carp "Value of surface ignored: specify either cr or surface";
+	    carp "Value of surface option ignored: specify only one of either context or surface";
 	}
 	if ($self->{verbose}) {
 	    debugmsg ("Using user-supplied context $self->{cr}");
@@ -61,6 +65,8 @@ sub new
     return $self;
 }
 
+# Make the Cairo context for our surface.
+
 sub make_cr
 {
     my ($self) = @_;
@@ -104,23 +110,189 @@ sub render
 	}
 	$p->parsefile ($file);
     }
+    $self->{depth} = 0;
+    $self->_render ($self->{top});
     return $self->{surface};
 }
 
+# Actually render
+
+sub _render
+{
+    my ($self, $element, $pattr) = @_;
+    my $tag = $element->{tag};
+    if ($tag eq 'defs') {
+	return;
+    }
+    my $attr = $self->_draw ($element, $pattr);
+    $self->{depth}++;
+    my $child = $element->{child};
+    for (@$child) {
+	$self->_render ($_, $attr);
+    }
+    $self->{depth}--;
+    $self->_draw_end ($element);
+}
+
+sub processUse
+{
+    my ($self, %attr) = @_;
+    my $id = $attr{'xlink:href'};
+    if (! $id) {
+	$id = $attr{href};
+    }
+    if (! $id) {
+	carp "No xlink:href/href in <use>";
+	return;
+    }
+    $id =~ s/^#//;
+    my $element = $self->get_id ($id);
+    if (! $element) {
+	carp "ID $id in use not found";
+	return;
+    }
+    my $cr = $self->{cr};
+    my $x = $attr{x};
+    my $y = $attr{y};
+    if (defined $x || defined $y) {
+	if (! defined $x) {
+	    $x = 0;
+	}
+	if (! defined $y) {
+	    $y = 0;
+	}
+	$self->msg ("Moving to $x $y");
+	$cr->save ();
+	$cr->translate ($x, $y);
+    }
+
+    $self->_render ($element, \%attr);
+
+    if (defined $x || defined $y) {
+	$cr->restore ();
+    }
+}
+
+sub _draw
+{
+    my ($self, $element, $pattr) = @_;
+    my $tag = $element->{tag};
+    $self->msg ("<$tag>");
+    my $attr = $element->{attr};
+
+    # %attr is a holder for inherited attributes. The inherited
+    # attributes are not written into $element->{attr} since it's
+    # possible that a <use> element might re-use the element, but want
+    # to give it different inherited attributes, so %attr does the job
+    # of keeping the actual element attributes (the ones written in
+    # the SVG itself) separate from the inherited (implicit)
+    # attributes.
+
+    my %attr = %$attr;
+
+    # This list of inherited things is guesswork so far, there is
+    # probably a list of what to copy somewhere but I haven't located
+    # it yet.
+
+    for my $key (qw!
+	fill
+	fill-opacity
+	fill-rule
+	opacity
+	stroke
+	stroke-dasharray
+	stroke-dashoffset
+	stroke-linecap
+	stroke-linejoin
+	stroke-miterlimit
+	stroke-opacity
+	stroke-width
+    !) {
+	if ($pattr->{$key} && ! $attr{$key}) {
+	    $attr{$key} = $pattr->{$key};
+	}
+    }
+
+    $self->do_svg_attr (%attr);
+    if ($tag eq 'svg' || $tag eq 'g') {
+	# These are non-rendering, i.e. don't result in visual output.
+	;
+    }
+    elsif ($tag eq 'path') {
+	$self->path (%attr);
+    }
+    elsif ($tag eq 'polygon') {
+	$self->polygon (%attr);
+    }
+    elsif ($tag eq 'line') {
+	$self->line (%attr);
+    }
+    elsif ($tag eq 'circle') {
+	$self->circle (%attr);
+    }
+    elsif ($tag eq 'ellipse') {
+	$self->ellipse (%attr);
+    }
+    elsif ($tag eq 'rect') {
+	$self->rect (%attr);
+    }
+    elsif ($tag eq 'polyline') {
+	$self->polyline (%attr);
+    }
+    elsif ($tag eq 'use') {
+	$self->processUse (%attr);
+    }
+    elsif ($tag eq 'defs') {
+	# Throw an exception. Arriving here is a bug, we should have
+	# stopped at _render. The <defs> element is processed and
+	# stored by the parser, but it is not used directly by the
+	# renderer. Its children are probably used by a <use> element.
+	confess "<defs> element reached";
+    }
+    else {
+	if ($self->{verbose}) {
+	    # There are probably many of these since this module is
+	    # not up to spec, so only complain if the user wants
+	    # "verbose" messages.
+	    carp "Unable to draw SVG element '<$tag>'";
+	}
+    }
+    $self->do_fill_stroke (\%attr);
+    return \%attr;
+}
+
+sub _draw_end
+{
+    my ($self, $element) = @_;
+    my $tag = $element->{tag};
+    $self->msg ("</$tag>");
+    # Only use the actual attributes, not the inherited ones, although
+    # the transform attribute is probably not inherited.
+    my $attr = $element->{attr};
+    if ($attr->{transform}) {
+	my $cr = $self->{cr};
+	$cr->restore ();
+    }
+}
+
 sub handle_end
 {
     # Last argument is $tag
     my ($self, $parser, undef) = @_;
     my $element = pop @{$self->{elements}};
     my $attr = $element->{attr};
-    $self->do_fill_stroke ($attr);
 }
 
-# <svg> tag seen
-
 sub svg
 {
     my ($self, %attr) = @_;
+
+    # Try to work out the height and width of the image. SVG is a very
+    # complicated format, so the height and width can be stored in
+    # multiple places.
+
+    my $min_x;
+    my $min_y;
     my $width;
     my $height;
     if ($attr{width}) {
@@ -131,20 +303,13 @@ sub svg
 	$height = $attr{height};
 	$height = svg_units ($height);
     }
-    if ($attr{fill}) {
-	$self->{fill} = $attr{fill};
-    }
-    if ($attr{stroke}) {
-	$self->{stroke} = $attr{stroke};
-    }
-
 
     # Use viewBox attribute
 
     if (! defined $width && ! defined $height) {
 	my $viewBox = $attr{viewBox} || $attr{viewbox};
 	if ($viewBox) {
-	    (undef, undef, $width, $height) = split /\s+/, $viewBox;
+	    ($min_x, $min_y, $width, $height) = split /\s+/, $viewBox;
 	}
     }
     my $surface = $self->{surface};
@@ -173,7 +338,65 @@ sub svg
 	$self->{surface} = $surface;
 	$self->make_cr ();
     }
-    $self->do_svg_attr (%attr);
+    if (defined $min_x && defined $min_y && ($min_x != 0 || $min_y != 0)) {
+	my $cr = $self->{cr};
+	$cr->translate (-$min_x, -$min_y);
+    }
+
+    my $svg = {
+	tag => 'svg',
+	attr => \%attr,
+	child => [],
+    };
+
+    # Store of ids of elements within the tree
+    $self->{ids} = {};
+    # Currently open tags (misnamed)
+    $self->{elements} = [$svg];
+    $self->{top} = $svg;
+}
+
+sub add_element
+{
+    my ($self, $tag, $attr) = @_;
+    my $element = {
+	tag => $tag,
+	attr => $attr,
+	child => [],
+    };
+    my $top = $self->{elements}[-1];
+    if (! $top) {
+	die "Empty stack";
+    }
+    push @{$top->{child}}, $element;
+    $element->{parent} = $top;
+    push @{$self->{elements}}, $element;
+    return $element;
+}
+
+# Store an ID so it can be retrieved later.
+
+sub store_id
+{
+    my ($self, $element, $id) = @_;
+    my $already = $self->{ids}{$id};
+    if ($already) {
+	carp "Duplicate id '$id' on element '<$already->{tag}>'";
+	return;
+    }
+    $self->{ids}{$id} = $element;
+}
+
+# Retrieve an element by ID. This is used by <use>.
+
+sub get_id
+{
+    my ($self, $id) = @_;
+    my $already = $self->{ids}{$id};
+    if ($already) {
+	return $already;
+    }
+    return undef;
 }
 
 # Start tag handler for the XML parser. This is private.
@@ -182,87 +405,25 @@ sub handle_start
 {
     my ($self, $parser, $tag, %attr) = @_;
 
-    my $parent = $self->{elements}[-1];
-    if ($parent) {
-	my $pattr = $parent->{attr};
-	for my $key (qw!
-	    fill
-	    stroke
-	    stroke-linecap
-	    stroke-linejoin
-	    stroke-width
-	!) {
-	    # So where were the spiders
-	    # While the fly tried to break our balls
-	    if ($pattr->{$key} && ! $attr{$key}) {
-		$attr{$key} = $pattr->{$key};
-	    }
-	}
-    }
-
-    my $element = {
-	tag => $tag,
-	attr => \%attr,
-    };
-    push @{$self->{elements}}, $element;
-
     if ($tag eq 'svg') {
 	$self->svg (%attr);
     }
-    elsif ($tag eq 'path') {
-	$self->path (%attr);
-    }
-    elsif ($tag eq 'polygon') {
-	$self->polygon (%attr);
-    }
-    elsif ($tag eq 'line') {
-	$self->line (%attr);
-    }
-    elsif ($tag eq 'circle') {
-	$self->circle (%attr);
-    }
-    elsif ($tag eq 'ellipse') {
-	$self->ellipse (%attr);
-    }
-    elsif ($tag eq 'rect') {
-	$self->rect (%attr);
-    }
-    elsif ($tag eq 'title') {
-	;
-    }
-    elsif ($tag eq 'g') {
-	$self->g (%attr);
-    }
-    elsif ($tag eq 'polyline') {
-	$self->polyline (%attr);
-    }
     else {
-	if ($self->{verbose}) {
-	    # There are probably many of these since this module is
-	    # not up to spec, so only complain if the user wants
-	    # "verbose" messages.
-	    carp "Unhandled SVG tag '$tag'";
+	my $element = $self->add_element ($tag, \%attr);
+	# I don't think svgs ids need to be stored here.
+	if ($attr{id}) {
+	    $self->store_id ($element, $attr{id});
 	}
     }
-
-    # http://www.princexml.com/doc/7.1/svg/
-    # g, rect, circle, ellipse, line, path, text, tspan
-    # Also "use" etc.
 }
 
-sub g
-{
-    my ($self, %attr) = @_;
-    # Group element
-}
-
-# Around the rugged rectangle the ragged rascals ran
+# Around the rugged rectangle the ragged rascals ran. "Polyfill" for
+# Cairo since it has no native rounded rectangles.
 
 sub rounded_rectangle
 {
     my ($self, %attr) = @_;
     my $cr = $self->{cr};
-    $self->do_svg_attr (%attr);
     # https://www.cairographics.org/samples/rounded_rectangle/
     my $x = svg_units ($attr{x});
     my $y = svg_units ($attr{y});
@@ -277,10 +438,11 @@ sub rounded_rectangle
 	    $rx = svg_units ($attr{rx});
 	}
     }
-#    my $ry;
-#    if ($attr{ry}) {
-#	$ry = svg_units ($attr{ry});
-#    }
+    # This is a kludge/hack at the moment.
+    #    my $ry;
+    #    if ($attr{ry}) {
+    #	$ry = svg_units ($attr{ry});
+    #    }
     $cr->new_sub_path ();
     $cr->arc ($x + $width - $rx, $y +           $rx, $rx, -pi/2,      0);
     $cr->arc ($x + $width - $rx, $y + $height - $rx, $rx,     0,   pi/2);
@@ -307,7 +469,6 @@ sub rect
 
     $cr->rectangle ($x, $y, $width, $height);
 
-    $self->do_svg_attr (%attr);
 }
 
 sub ellipse
@@ -333,7 +494,6 @@ sub ellipse
 
     $cr->restore ();
 
-    $self->do_svg_attr (%attr);
 }
 
 sub circle
@@ -349,8 +509,6 @@ sub circle
     # Render it.
 
     $cr->arc ($cx, $cy, $r, 0, 2*pi);
-
-    $self->do_svg_attr (%attr);
 }
 
 sub split_points
@@ -380,7 +538,6 @@ sub polygon
 	$cr->line_to ($x, $y);
     }
     $cr->close_path ();
-    $self->do_svg_attr (%attr);
 }
 
 sub polyline
@@ -401,7 +558,6 @@ sub polyline
 	$x = pop @points;
 	$cr->line_to ($x, $y);
     }
-    $self->do_svg_attr (%attr);
 }
 
 sub path
@@ -433,19 +589,28 @@ sub path
 	if ($key eq lc $key) {
 	    # This is a bug, "extract_path_info" above should never
 	    # return a lower-case key, which means a relative path.
-	    die "Path parse conversion to absolute failed";
+	    confess "Path parse conversion to absolute failed";
 	}
 
 	if ($key eq 'S') {
 	    # This is a bug, "extract_path_info" above should never
 	    # return a shortcut key, they should have been converted
 	    # to C keys.
-	    die "Path parse conversion to no shortcuts failed";
+	    confess "Path parse conversion to no shortcuts failed";
 	}
 	if ($key eq 'M') {
 	    # Move to
 	    $cr->new_sub_path ();
 	    $cr->move_to (@{$element->{point}});
+
+	    # This is debugging code from the changeover to two-stage
+	    # rendering.
+
+	    # $self->msg ("Move to @{$element->{point}}");
+	    # print $cr->status (), "\n";
+	    # my @p1 = $cr->get_current_point ();
+	    # $self->msg ("Move to @p1");
+
 	}
 	elsif ($key eq 'L') {
 	    $cr->line_to (@{$element->{point}});
@@ -482,7 +647,6 @@ sub path
 	    carp "Unknown SVG path key '$key': ignoring";
 	}
     }
-    $self->do_svg_attr (%attr);
 }
 
 # This is a Perl translation of 
@@ -665,7 +829,7 @@ sub quadbez
     if (! $cr->has_current_point ()) {
 	# This indicates a bug has happened, because there is always a
 	# current point when rendering an SVG path.
-	die "Invalid drawing of quadratic bezier without a current point";
+	confess "Invalid drawing of quadratic bezier without a current point";
     }
 
     my @p1 = $cr->get_current_point ();
@@ -693,7 +857,6 @@ sub line
     my $cr = $self->{cr};
     $cr->move_to ($attr{x1}, $attr{y1});
     $cr->line_to ($attr{x2}, $attr{y2});
-    $self->do_svg_attr (%attr);
 }
 
 my %units = (
@@ -768,23 +931,115 @@ sub do_svg_attr
     if ($linejoin) {
 	$cr->set_line_join ($linejoin);
     }
-    my $fill = $attr{fill};
-    if ($fill) {
-	$fill = trim ($fill);
+    my $transform = $attr{transform};
+    if ($transform) {
+	$cr->save ();
+	$self->do_transforms (%attr);
     }
-    if (! $fill) {
-	my $svgfill = $self->{svg}{fill};
-	if ($svgfill) {
-	    $svgfill = trim ($svgfill);
+}
+
+# The reason this is as complicated as it is is because SVG accepts
+# not only commas and/or spaces as separators, but also things like
+# -30-40 as two numbers in transform arguments. Did this
+# "optimization" by SVG designers really do anything useful?
+
+my $sep = qr!(?:\s+|\s*,\s*)!;
+my $num = qr![-0-9\.]+!;
+my $sepnum = qr!(?:$sep$num|$sep?-$num)!;
+
+sub sepnum
+{
+    my ($sepnum) = @_;
+    $sepnum =~ s!^$sep!!;
+    return $sepnum;
+}
+
+sub do_transforms
+{
+    my ($self, %attr) = @_;
+    my $cr = $self->{cr};
+    # Transformers - robots in disguise
+    my $transform = $attr{transform};
+    while ($transform =~ /((?:translate|scale|rotate|matrix)\s*\([^\)]*\))/g) {
+	my $change = $1;
+	if ($change =~ /translate\s*\(\s*($num)($sepnum)\s*\)/) {
+	    my $x = $1;
+	    my $y = sepnum ($2);
+	    $self->msg ("translate ($x, $y)");
+	    $cr->translate ($x, $y);
+	    next;
+	}
+	if ($change =~ /scale\s*\(\s*($num)(?:$sep($num))?\s*\)/) {
+	    my $x = $1;
+	    my $y = $2;
+	    if (defined $y) {
+		$y = sepnum ($y);
+	    }
+	    else {
+		# scale may take one argument
+		$y = $x;
+	    }
+	    $self->msg ("scale ($x, $y)");
+	    $cr->scale ($x, $y);
+	    next;
+	}
+	if ($change =~ /
+	    rotate\s*\(
+	    \s*($num)\s*
+	    (?:($sepnum)($sepnum))?
+	    \s*\)/x) {
+	    my $angle = $1;
+	    my $x = $2;
+	    my $y = $3;
+	    my $trans;
+	    if (defined $x && defined $y) {
+		$x = sepnum ($x);
+		$y = sepnum ($y);
+		$trans = 1;
+	    }
+	    if ($trans) {
+		$cr->translate ($x, $y)
+	    }
+	    $cr->rotate (deg2rad ($angle));
+	    if ($trans) {
+		$cr->translate (-$x, -$y)
+	    }
+	    if ($trans) {
+		$self->msg ("rotate $angle around $x $y");
+	    }
+	    else {
+		$self->msg ("rotate $angle !");
+	    }
+	    next;
+	}
+	if ($change =~ m!
+	    matrix\s*
+	    \(\s*
+	    ($num)
+	    ($sepnum)
+	    ($sepnum)
+	    ($sepnum)
+	    ($sepnum)
+	    ($sepnum)
+	    \s*\)
+	    !x) {
+	    my @nums = ($1, $2, $3, $4, $5, $6);
+	    @nums = map {sepnum ($_)} @nums;
+	    $self->msg ("Matrix @nums");
+	    my $m = Cairo::Matrix->init (@nums);
+	    my $matrix = $cr->get_matrix ();
+	    $matrix = $matrix->multiply ($m);
+# I'm not yet sure how to implement the translate part.
+	    #$matrix =
+# $matrix->translate (-$nums[4]/2, $nums[5]);
+	    $cr->set_matrix ($matrix);
+	    next;
 	}
     }
-    my $stroke = $attr{stroke};
-    if ($stroke) {
-	trim ($stroke);
-    }
-    my $fill_opacity = $attr{'fill-opacity'};
-    # Not sure how to handle this yet.
-    #    $self->do_fill_stroke ($cr, $fill, $stroke);
+    # $transform = trim ($transform);
+    # if ($transform) {
+    # 	warn "Unhandled '$transform'";
+    # }
 }
 
 sub do_fill_stroke
@@ -793,38 +1048,49 @@ sub do_fill_stroke
     my $cr = $self->{cr};
     my $fill = $attr->{fill};
     my $stroke = $attr->{stroke};
+    # These can be undefined
+    my $fill_opacity = $attr->{'fill-opacity'};
+    my $stroke_opacity = $attr->{'stroke-opacity'};
+    my $opacity = $attr->{opacity};
+    if (defined $opacity && ! defined $fill_opacity) {
+	$fill_opacity = $opacity;
+    }
+    if (defined $opacity && ! defined $stroke_opacity) {
+	$stroke_opacity = $opacity;
+    }
 
     if ($fill && $fill ne 'none') {
 	if ($stroke && $stroke ne 'none') {
-	    $self->set_colour ($fill);
+	    $self->set_colour ($fill, $fill_opacity);
 	    $cr->fill_preserve ();
 	    $self->msg ("Filling with $fill");
-	    $self->set_colour ($stroke);
+	    $self->set_colour ($stroke, $stroke_opacity);
 	    $cr->stroke ();
 	    $self->msg ("Stroking with $stroke");
 	}
 	else {
-	    $self->set_colour ($fill);
+	    $self->set_colour ($fill, $fill_opacity);
 	    $self->msg ("Filling with $fill");
 	    $cr->fill ();
 	}
     }
     elsif ($stroke && $stroke ne 'none') {
-	$self->set_colour ($stroke);
+	$self->set_colour ($stroke, $stroke_opacity);
+	$self->msg ("Stroking with $stroke");
 	$cr->stroke ();
     }
     elsif (! $fill && ! $stroke) {
 	$self->msg ("Filling with black");
 	# Fill with black seems to be the default.
-	$self->set_colour ('#000000');
+	$self->set_colour ('#000000', $fill_opacity);
 	$cr->fill ();
     }
 }
 
 # Graphics::ColorNames::WWW for some reason returns these as integers
-# with the R, G, and B components multiplied together, so to use that
-# module we would need to then divide the numbers to get the R, G and
-# B values back. It was easier just to copy and paste.
+# with the R, G, and B components multiplied by factors of 256, so to
+# use that module we would need to then divide the numbers to get the
+# R, G and B values back. It was easier just to copy and paste.
 
 my %color2rgb = (
     'aliceblue'         => [240, 248, 255],
@@ -988,7 +1254,7 @@ sub name2colour
 
 sub set_colour
 {
-    my ($self, $colour) = @_;
+    my ($self, $colour, $opacity) = @_;
     my $cr = $self->{cr};
     # Hex digit
     my $h = qr/[0-9a-f]/i;
@@ -1009,7 +1275,16 @@ sub set_colour
     else {
 	@c = name2colour ($colour);
     }
-    $cr->set_source_rgb (@c);
+    if (defined $opacity) {
+	if ($opacity > 1 || $opacity < 0) {
+	    carp "Opacity value $opacity out of bounds";
+	    $opacity = 1;
+	}
+	$cr->set_source_rgba (@c, $opacity);
+    }
+    else {
+	$cr->set_source_rgb (@c);
+    }
 }
 
 sub surface
@@ -1042,6 +1317,7 @@ sub msg
     if (! $self->{verbose}) {
 	return;
     }
+    print "  " x $self->{depth};
     print "$msg\n";
 }
 
